#!/usr/bin/perl -w
my $version = "1.0";	

### USER CONFIGURATION **#######################################################

my $DEBUG      = 0;			# Use EXTREMELY verbose error messages
my $NNTPAUTH   = "$ENV{'HOME'}/.nntpauth";
my $KIBOZE     = "$ENV{'HOME'}/kiboze";	# Base directory to store files in
my $ARCHIVES   = "$KIBOZE/archives";	# Where to keep the archives
my $CACHEFILE  = "$KIBOZE/messages";	# Where to keep the message-ID cache
my $LOCALNAME  = "archive";		# Local "machine" name, as you want it 
					#   to appear in the Xref and Path 
my $MAXARTS    = 1000;			# Maximum number of articles 
					#   to download per invocation
my %SERVERS    = (			# News servers and their Newsrcs
	$ENV{'NNTPSERVER'}	=>	"$KIBOZE/newsrc",
		 );

################################################################################

=head1 NAME

kiboze.pl - a news archiving 'bot

=head1 SYNOPSIS

kiboze.pl [-hvd] [B<-V>] [B<-Q>] [B<-n> newsrcfile] [B<-s> newsserver]

=head1 DESCRIPTION

This script connects to a news server (or set of servers) and archives the
newsgroups specified in a newsrc file.  This can be used to save entire
newsgroups for later searches,  just message archiving.  

=head1 USAGE

Without any options, kiboze.pl will connect to each of the news servers
specified in turn, and attempt to download up to 1000 articles (or whatever 
the current value of $MAXARTS is).  

=over 4

=item -h

Prints a short help message and exits.

=item -v

Prints the version number and exits.

=item -d

Prints out the script defaults and exits.

=item -V

Verbose mode.  Prints out some basic status messages when they are
necessary.  More information is available in debugging mode, but you
probably don't want to read it.

=item -Q

Quiet mode.  Never prints anything out.  This is handy if you're putting
the script into a cron job and don't want to look over any logs; on the
other hand, most of what it blocks is error messages anyway, so you might
want to have them in your mailbox.  

=item -s F<newsserver>

Specifies which news server to connect to.  This may be used to connect to
just one server out of a set.  If combined with the B<-n> flag, you can also
connect to servers not present in the script.  

=item -n F<newsrcfile>

Specifies which newsrc to use.  If this newsrc is already used by a
server, then it uses that information for the server; otherwise, the B<-s>
flag is required.  

=back

=head1 REQUIREMENTS

Requires Perl 5.004 or better, News::Newsrc, and News::Article.  All other
requirements should be included automatically with Perl5.

=cut

# Version History
# 1.0 Thu Nov 16 15:05:08 CST 2000
#   Made an install script, fixed the documentation a bit, started
#   running with warnings, submitted it to freshmeat.
# 0.6b Wed Aug 30 10:21:15 CDT 2000
#   Added basic support for NNTPAUTH stuff - it works, at least.  Not
#   documented yet, and it probably ought to be standardized.
# 0.51b Thu Feb  3 12:10:46 CST 2000
#   Added "use Net::NNTP", since it's not included with News::Article
#   anymore.
# 0.5b Tue Dec 21 17:19:57 CST 1999
#   First version that I actually wrote up some documentation and such for.
#   It's been working (and running) for a long time before this, but it
#   was about time I got around to making it useful.  Let's see how long
#   it'll take me to start distributing it.

umask 022;		# Created files should be permissions 777
require 5.004;		# Require at least Perl 5.004
use strict;		# Good programming is our friend

use Getopt::Std;	# Command line functions
use vars qw($opt_h $opt_v $opt_Q $opt_V $opt_s $opt_n $opt_d);


# Command-line configuration
$0 =~  s%.*/%%;			# Clean up the program name
getopts('hvVqds:n:');

&Usage 		if $opt_h;	# Print a help message and exit
&Version 	if $opt_v;	# Print a version message and exit
&Default	if $opt_d;	# Prints out the defaults

my $VERBOSE      = $opt_V;	# Give verbose messages
my $QUIET        = $opt_Q;	# Give no output at all

my @SERVERS;
if ($opt_s) {   # Just use the server given in -s 
  push (@SERVERS, $opt_s);
  if ($opt_n) {  # Use the given newsrc for this server
    $SERVERS{$opt_s} = $opt_n;    
  } elsif ( defined($SERVERS{$opt_s}) ) {	# Just use this server
    # Nothing to do here.  
  } else {	 # Not enough information to continue, so exit
    Exit('CONFIG', "Can't find newsrc file for server $opt_s");
  }
} else {	# Use all of the servers in %SERVERS
  if ($opt_n) {	# Limit to just the server with the given newsrc

    # Get the whole name out of $opt_n
    my $newsrc = $opt_n;
       $newsrc = "$ENV{'PWD'}/$newsrc" unless ($opt_n =~ m%^/%);

    foreach (keys %SERVERS) {
      push (@SERVERS, $_) if ($newsrc eq $SERVERS{$_});
    }  

    if ( scalar(@SERVERS) > 1 ) { 	# Check the matches
      Exit('CONFIG', "Too many matches on newsrc $opt_n - @SERVERS");
    } elsif ( scalar(@SERVERS) == 0 ) {
      Exit('CONFIG', "No servers found for newsrc $opt_n");
    }
  } else {
    @SERVERS = keys(%SERVERS);
  }
}

use News::Newsrc;	# News functions
use News::Article;
use Net::NNTP;

use Fcntl;		# Database functions
use SDBM_File;

### Main Program ###

Exit('CONFIG', "No servers to scan") unless ( scalar(@SERVERS) >= 1 );
Exit('CONFIG', "LOCALNAME not set")  unless $LOCALNAME;

# Load up the cache file
print "Loading up the Message-ID cache file $CACHEFILE..." if ($DEBUG);

my %ID_Cache;
tie %ID_Cache, "SDBM_File", $CACHEFILE, O_CREAT|O_RDWR, 0755
  or Exit('CONFIG', "Couldn't open Message-ID Cache file $CACHEFILE");

print "...done\n" if ($DEBUG);
  
my ($group, $server);
foreach $server (@SERVERS) {
  my $newsrcfile = $SERVERS{$server};
  print "$server - loading $newsrcfile\n" if $VERBOSE;
  # Load the newsrc file
  my $Newsrc = new News::Newsrc;
   unless ( $Newsrc->load($newsrcfile) ) {
     warn "Couldn't open $newsrcfile for server $server\n" unless $QUIET; 
     next; 
   }

  # Load $NNTPAUTH
  my ($nntpuser, $nntppass);
  if (-r $NNTPAUTH) {
    open (CONFIG, "<$NNTPAUTH") or warn "Couldn't open $NNTPAUTH";
    while (<CONFIG>) {
      chomp;
      if (/^\s*(\S+)\s+(\S+)\s+(\S+)\s*$/) {
        my $serv = $1;  my $user = $2;  my $pass = $3;
        $nntpuser = $user if ($serv =~ /^$server$/i);
        $nntppass = $pass if ($serv =~ /^$server$/i);
      }
    }
    $nntpuser ||= "";  $nntppass ||= "";
    print "Server: $server  User: $nntpuser  Pass: $nntppass\n" if $VERBOSE;
    close(CONFIG);
  }

  # Create the directory structure from $ARCHIVES and the group names
  foreach $group ($Newsrc->groups) {
    my $directory = $group =~ s%\.%/%g; # .'s become directory delimiters
    Exit('CONFIG', "Couldn't create directory $ARCHIVES/$group") 
      if ( mkdir_full("$ARCHIVES/$group") );
  }
  
  # Connect to the NNTP Server
  my $NNTP = Net::NNTP->new($server);
  $NNTP->authinfo($nntpuser, $nntppass) if defined($nntppass);
  Exit('SERVER', "Couldn't connect to server $server") unless $NNTP;
  
  my $count = 0;
  foreach $group ($Newsrc->groups) {
    my ($articles, $firstnum, $lastnum, $name) = $NNTP->group($group);
    next unless $name;		
    
    print "$name: articles $firstnum - $lastnum exist on server\n" if $VERBOSE; 
  
    my $i;
    foreach ($i = $firstnum; $i <= $lastnum; $i++) {
      last if ($count >= $MAXARTS);		
      next if $Newsrc->marked($group, $i);	# Next if it's marked

      # Get the article and its information
      my $article = News::Article->new( $NNTP->article($i) );
      
      # If there is an article and it's not already in the cache, save it
      if ($article && !$ID_Cache{ $article->header('message-id') }) {
        my $messageid = $article->header('message-id');
        my $file = save_article($article, $ARCHIVES, $Newsrc);
           $file =~ s%^$ARCHIVES/?%%;	
        if ($file) {	# Did it save?
          $ID_Cache{ $messageid } = $file;	# Update the cache
          $count++;				# Update the count
          print "Saved $messageid as $file\n" if $DEBUG;
        } else {
          print "Couldn't save article $messageid\n" if $DEBUG;
        }
      } elsif ( $article ) {	# It was already in the cache
        my $messageid = $article->header('message-id');
        my $path      = $ID_Cache{ $messageid };
        print "Article $messageid already archived - $path\n" if $DEBUG;
      } else {			# There was no article, for whatever reason
        print "Didn't actually get an article, skipping...\n" if $DEBUG; 
      } 
      $Newsrc->mark($group, $i);
    }
    $Newsrc->save;
  }
  print "$count articles archived for server $server\n" if $VERBOSE;
  
  # Close the NNTP Connection
  $NNTP->quit;
}

untie %ID_Cache;


### Function Calls ###

## mkdir_full ( DIRECTORY )
# Takes a fully qualified directory name and creates it.  Returns undef if
#   successful, or an error message otherwise.

sub mkdir_full {
  my $directory = shift;
  my @toprocess;

  # Figure out which directory does exist and maintain a list of those 
  #   elements after it, so that we can make the directories on the way

  until (-d $directory) {	
    $directory =~ s%/$%%;	# Remove trailing /'s
    my ($base, $dir) = $directory =~ m%^(.*)/([^/]*)$%;
    return "Bad directory name: $directory" unless $dir;
    push (@toprocess, $dir);
    $directory = $base;
  }
  
  return undef unless @toprocess; 	# We're done if it's empty

  foreach (reverse @toprocess) { 
    return "Can't create directory under $directory" unless (-w $directory);
    mkdir("$directory/$_", 0777) 	# Create the directory with said umask
	or return "Couldn't create directory $directory/$_" ;
    print "Created directory $directory/$_\n" if $DEBUG;
    $directory = "$directory/$_";
  }
  return undef;
}

## save_article ( ARTICLE, BASEDIR, NEWSRC )
# Saves the given ARTICLE off of BASEDIR.  NEWSRC is used to figure out
#   which groups have to be checked out - this makes this not as generic
#   as it could be, but it'll do.  

sub save_article {
  my $article = shift;
  my $basedir = shift;
  my $Newsrc  = shift;
  my @groups = split('\s*,\s*', $article->header('newsgroups'));

  # Create a new Path header to reflect the new file location
  my $newpath = join('!', $LOCALNAME || 'localhost', $article->header('Path'));
  $article->set_headers('Path', $newpath);

  # Figures out where to save the article
  my ($file, $group);
  my %files;
  foreach $group (@groups) {
    next unless $Newsrc->subscribed($group);	
    my $dir = $group;		
       $dir =~ s%\.%/%g;
    $files{$group} = _getfilename("$basedir/$dir");
  }
  
  # Make a new Xref header
  my $newxref = "$LOCALNAME";
  foreach (keys %files) {
    $newxref .= " $_:$files{$_}";
  }
  $article->set_headers('Xref', $newxref);

  # Creates the necessary files
  foreach $group (keys %files) {
    my $dir = $group; 
       $dir =~ s%\.%/%g;
    if ($file) {	# Already exists, just make a symlink
      my $link = "$basedir/$dir/$files{$group}";
      symlink ($file, $link);
    } else {		# Doesn't exist yet, create a new file
      $file = "$basedir/$dir/$files{$group}";
      open (FILE, ">$file") or Exit('CONFIG', "Couldn't create file $file!");
      $article->write(\*FILE);
      close (FILE);
    }
  }
  return $file;
}

## _getfilename ( DIRECTORY )
# Gets the next file name for the given DIRECTORY, which is the highest
#   number in the group + 1 (or 1 if the directory is empty).  This could 
#   probably be more efficient if we had some extra datafiles, but we 
#   don't, so it'll do.

sub _getfilename {
  my $directory = shift;
  opendir (DIRECTORY, $directory) or return undef;
  my $count = 1;
  foreach (grep /^(\d+)$/, readdir DIRECTORY) {
    $count = $_ + 1 if ($_ >= $count);
  }
  closedir (DIRECTORY);
  return $count;
}

## Usage ()
# Prints out a short help file and exits.

sub Usage {
  warn <<EOM;

$0 v$version
A newsgroup archiver
Usage: $0 [-hvdVQ] [-s newsserver] [-n newsrcfile] 

$0 connects to a news server (or set of servers) and archives the specified 
newsgroups from a newsrc file (or files).  The defaults are stored in the 
script.  For more information, refer to the manual page.

	-h		Prints this message and exits.
	-v		Prints the version number and exits.
	-d		Prints out the script defaults and exits.  
	-V		Verbose mode.  Gives some basic status messages.
	-Q		Quiet mode.  Doesn't print any messages.  Useful for
			  cron jobs, but otherwise not recommended.
	-s newsserver   Specifies which server to connect to, instead of 
			  the whole list.  May be combined with -n to connect
			  to servers not present in the script.
	-n newsrcfile	Specifies which newsrc to use.  Requires that the
			  file is already mentioned in the script, or that
			  the -s flag is used.

EOM

  Exit();
}

## Version ()
# Prints out the version number and exits

sub Version {
  warn "$0 v$version\n";
  Exit();
}

## Default()
# Prints out the program defaults and exits

sub Default {
  unless ($QUIET) {
    warn "Debugging mode on\n" if $DEBUG;
    warn "Base directory:		$KIBOZE\n";
    warn "Archives:		$ARCHIVES\n";
    warn "Cachefile:		$CACHEFILE\n";
    warn "Local name:		$LOCALNAME\n";
    warn "Maximum articles: 	$MAXARTS\n";
    warn "\n       Servers and Newsrcs\n";
    foreach (keys %SERVERS) {
      warn "  Server: $_\n  Newsrc: $SERVERS{$_}\n\n";
    }
  }
  Exit();
}

## Exit ( CODE, REASON )
# Exits the program cleanly with a proper error message

my %ErrorCodes = (
	'CONFIG'	=> 	1,
	'SERVER'	=>	2,
		 );

sub Exit {
  my ($code, @reason) = @_;
  warn "@reason\n" if (!$QUIET && @reason);
  exit $ErrorCodes{$code};
}

=head1 TODO

Make some default newsrc files if one doesn't exist.  Start using an 
active-file type dealie, to increase efficiency by a whole lot and allow
the various servers to work together.  Make some more scripts to search 
the cache file by message-id.  Add locking stuff.  

=head1 AUTHOR

Tim Skirvin <tskirvin@killfile.org>

=head1 COPYRIGHT

Copyright 1999-2000 Tim Skirvin <tskirvin@killfile.org>

This code may be used and/or distributed under the same terms as Perl
itself.  Oh, and it's beta code, so use it at your own risk.

=cut

